# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Documents are wrappers around different forms of snapshots that may be
committed to an archive.
"""

from abc import ABCMeta, abstractmethod


class Document(metaclass=ABCMeta):
    """The document interface provides access to a document reader. The reader
    is expected to give access to a document that is sorted in ascending order
    of the row key that is used by an archive during merge.
    """
    def __init__(self, columns):
        """Initialize the object properties. The abstract document class
        maintains a list of column names in the document schema. Columns may
        either be represented by strings or by instances of the Column class.

        Parameters
        ----------
        columns: list
            List of column names. The number of values in each document row is
            expected to be the same as the number of columns and the order of
            values in each row is expected to correspond to their respective
            column in this list.
        """
        self.columns = columns

    @abstractmethod
    def close(self):  # pragma: no cover
        """Signal that the archive merge is done reading the document. Any
        resources (e.g., temporary files) that were created for the document
        can be released.
        """
        raise NotImplementedError()

    @abstractmethod
    def partial(self, reader):  # pragma: no cover
        """Return a copy of the document that provides access to the set of
        rows such that the document is considered a partial document. In a
        partial document the rows are aligned with the position of the
        corresponding row in the document origin. The original row order is
        accessible via the given row position reader.

        Parameters
        ----------
        reader: histore.archive.reader.RowPositionReader
            Reader for row (key, position) tuples from the original
            snapshot version.

        Returns
        -------
        histore.document.base.Document
        """
        raise NotImplementedError()

    @abstractmethod
    def reader(self, schema):  # pragma: no cover
        """Get reader for data frame rows ordered by their row identifier. In
        a partial document the row positions that are returned by the reader
        are aligned with the positions of the corresponding rows in the
        document of origin.

        Parameters
        ----------
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the column list of this document (corresponding to
            their position in the list). The schema columns provide the unique
            column identifier that are required by the document reader to
            generate document rows. An error is raised if the number of
            elements in the schema does not match the number of columns in the
            data frame.

        Returns
        -------
        histore.document.reader.DocumentReader

        Raises
        ------
        ValueError
        """
        raise NotImplementedError()



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Iterator for document rows. Provides a reader that allows to iterate over
the rows in a data frame sorted by the row key.
"""

from abc import ABCMeta, abstractmethod


class DocumentReader(metaclass=ABCMeta):
    """Reader for rows in a document. Reads rows in order defined by the row
    key that is used for merging the document.
    """
    def __iter__(self):
        """Make the reader instance iterable by returning a generator that
        yields all rows.

        Returns
        -------
        Generator
        """
        return row_stream(self)

    @abstractmethod
    def has_next(self):  # pragma: no cover
        """Test if the reader has more rows to read. If True the next() method
        will return the next row. Otherwise, the next() method will return
        None.

        Returns
        -------
        bool
        """
        raise NotImplementedError()

    @abstractmethod
    def next(self):  # pragma: no cover
        """Read the next row in the document. Returns None if the end of the
        document has been reached.

        Returns
        -------
        histore.document.row.DocumentRow
        """
        raise NotImplementedError()


# -- Helper Methods -----------------------------------------------------------

def row_stream(reader):
    """Geterator that yields all rows in a document reader.

    Parameters
    ----------
    reader: histore.document.reader.DocumentReader
        Document reader over which we are iterating.

    Returns
    -------
    histore.document.row.DocumentRow
    """
    while reader.has_next():
        yield reader.next()



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Columns in histore data frames have a unique identifier and a column name.
The column class extends the Python String class to be able to be used as a
column value in a Pandas data frame.
"""


class Column(str):
    """Columns in openclean data frames are subclasses of Python strings that
    contain a unique column identifier. This implementation is based on:
    https://bytes.com/topic/python/answers/32098-my-experiences-subclassing-string

    The order of creation is that the __new__ method is called which returns
    the object then __init__ is called.
    """
    def __new__(cls, colid, name, *args, **keywargs):
        """Initialize the String object with the given column name. Ignore the
        column identifier.

        Parameters
        ----------
        colid: int
            Unique column identifier
        name: string
            Column name
        """
        return str.__new__(cls, str(name))

    def __init__(self, colid, name):
        """Initialize the unique column identifier. The column name has already
        been initialized by the __new__ method that is called prior to the
        __init__ method.

        Parameters
        ----------
        colid: int
            Unique column identifier
        name: string
            Column name
        """
        self.colid = colid


# -- Helper methods -----------------------------------------------------------


def column_index(schema, columns):
    """Get the list of column index positions in a given schema (list of
    column names). Columns are either specified by name or by index position.

    The result is a list of column index positions.

    Raises errors if invalid columns positions or unknown column names are
    provided.

    Parameters
    ----------
    schema: list(string)
        List of column names.
    columns: list(int or str)
        List of column index positions or column names.

    Returns
    -------
    (list, list)

    Raises
    ------
    ValueError
    """
    # Ensure that columns is a list.
    columns = columns if isinstance(columns, list) else [columns]
    result = list()
    for col in columns:
        if isinstance(col, int):
            # Raise value error if the specified index is invalid
            if col >= len(schema):
                raise ValueError("invalid index position '%d'" % (col))
            colidx = col
        else:
            colidx = -1
            for i, name in enumerate(schema):
                if name == col:
                    colidx = i
                    break
            # Raise value error if the column name is unknown
            if colidx == -1:
                raise ValueError('unknown column name {}'.format(col))
        result.append(colidx)
    return result



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Factory pattern for archives that are maintained on disk."""

import json
import os
import shutil

from histore.archive.base import PersistentArchive
from histore.archive.manager.base import ArchiveManager
from histore.archive.manager.descriptor import ArchiveDescriptor

import histore.config as config
import histore.util as util


class PersistentArchiveManager(ArchiveManager):
    """The persistent archive manager maintains a set of archives on disk. The
    list of archive descriptors is also maintained on disk as a Json file.
    """
    def __init__(self, basedir=None):
        """Initialize the base directory under which all archives are stores in
        individual sub-folders. If the base directory is not given the value
        will be read from the environment variable HISTORE_BASEDIR or the
        default value $HOME/.histore.

        Read the Json file containing the archive descriptors (if it exist).

        Parameters
        ----------
        basedir: string
            Path to dorectory on disk where archives are maintained.
        """
        if basedir is None:
            basedir = config.BASEDIR()
        self.basedir = util.createdir(basedir)
        # Initialize path to file that maintains archive descriptors.
        self.descriptorfile = os.path.join(self.basedir, 'archives.json')
        # Initialize the internal cache of archive descriptors
        self._archives = dict()
        if os.path.isfile(self.descriptorfile):
            with open(self.descriptorfile, 'r') as f:
                doc = json.load(f)
            for obj in doc:
                descriptor = ArchiveDescriptor(obj)
                self._archives[descriptor.identifier()] = descriptor

    def archives(self):
        """Get dictionary of archive descriptors. The returned dictionary maps
        archive identifier to their descriptor.

        Returns
        -------
        dict
        """
        return self._archives

    def create(self, name=None, description=None, primary_key=None):
        """Create a new archive object.

        Parameters
        ----------
        name: string, default=None
            Descriptive name that is associated with the archive.
        description: string, default=None
            Optional long description that is associated with the archive.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for rows in the
            archive.

        Returns
        -------
        histore.archive.manager.descriptor.ArchiveDescriptor
        """
        # Create the descriptor for the new archive.
        descriptor = ArchiveDescriptor.create(
            name=name,
            description=description,
            primary_key=primary_key
        )
        identifier = descriptor.identifier()
        primary_key = descriptor.primary_key()
        # Write list of archive descriptors.
        self._archives[identifier] = descriptor
        with open(self.descriptorfile, 'w') as f:
            json.dump([d.doc for d in self._archives.values()], f)
        return descriptor

    def delete(self, identifier):
        """Delete the archive with the given identifier.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        """
        if self.contains(identifier):
            shutil.rmtree(os.path.join(self.basedir, identifier))
            del self._archives[identifier]
            with open(self.descriptorfile, 'w') as f:
                json.dump([d.doc for d in self._archives.values()], f)

    def get(self, identifier):
        """Get the archive that is associated with the given identifier. Raises
        a ValueError if the identifier is unknown.

        Parameters
        ----------
        identifier: string
            Unique archive identifier

        Returns
        -------
        histore.archive.vase.Archive

        Raises
        ------
        ValueError
        """
        if identifier not in self._archives:
            raise ValueError('unknown archive {}'.format(identifier))
        archdir = os.path.join(self.basedir, identifier)
        primary_key = self._archives[identifier].primary_key()
        return PersistentArchive(basedir=archdir, primary_key=primary_key)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Manager implementation for archives that are maintained in main memory."""

from histore.archive.base import Archive
from histore.archive.manager.base import ArchiveManager
from histore.archive.manager.descriptor import ArchiveDescriptor


class VolatileArchiveManager(ArchiveManager):
    """The volatile archive manager maintains a set of archives in main memory.
    """
    def __init__(self):
        """Initialize the dictionaries that maintain archive descriptors and
        archive objects.
        """
        self._archives = dict()
        self._descriptors = dict()

    def archives(self):
        """Get dictionary of archive descriptors. The returned dictionary maps
        archive identifier to their descriptor.

        Returns
        -------
        dict
        """
        return self._descriptors

    def create(self, name=None, description=None, primary_key=None):
        """Create a new volatiole archive object.

        Parameters
        ----------
        name: string, default=None
            Descriptive name that is associated with the archive.
        description: string, default=None
            Optional long description that is associated with the archive.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for rows in the
            archive.

        Returns
        -------
        histore.archive.manager.descriptor.ArchiveDescriptor
        """
        # Create the descriptor for the new archive.
        descriptor = ArchiveDescriptor.create(
            name=name,
            description=description,
            primary_key=primary_key
        )
        identifier = descriptor.identifier()
        archive = Archive(primary_key=descriptor.primary_key())
        self._archives[identifier] = archive
        self._descriptors[identifier] = descriptor
        return descriptor

    def delete(self, identifier):
        """Delete the archive with the given identifier.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        """
        if self.contains(identifier):
            del self._archives[identifier]
            del self._descriptors[identifier]

    def get(self, identifier):
        """Get the archive that is associated with the given identifier. Raises
        a ValueError if the identifier is unknown.

        Parameters
        ----------
        identifier: string
            Unique archive identifier

        Returns
        -------
        histore.archive.vase.Archive

        Raises
        ------
        ValueError
        """
        if identifier not in self._archives:
            raise ValueError('unknown archive {}'.format(identifier))
        return self._archives[identifier]



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Base classes for row key values."""

from abc import ABCMeta
from functools import total_ordering

import numpy as np


class KeyValue(metaclass=ABCMeta):
    """Abstract base class for key values. We currently distinguish four types
    of key values: numeric, string, None, and new rows. The distinction between
    None and new rows is of importance for documents that are keyed by a list
    of numeric row index values. In that case, non-numeric or negative values
    are considered new rows that are assigned row identifier during the merge
    process.

    Supclasses primarily implement comparison operators for sorting. The
    general sort order is: numeric, string, null, new
    """
    def __init__(self, value=None):
        """Initialize the actual key value.

        Parameters
        ----------
        value: scalar, default=None
            This is the actual key value for a row.
        """
        self.value = value

    def is_new(self):
        """Test if the key represents a new row.

        Returns
        -------
        bool
        """
        return isinstance(self, NewRow)

    def is_null(self):
        """Test if the key represents a null value (None).

        Returns
        -------
        bool
        """
        return isinstance(self, NullKey)

    def is_number(self):
        """Test if the key is a numeric value.

        Returns
        -------
        bool
        """
        return isinstance(self, NumberKey)

    def is_string(self):
        """Test if the key is a string value.

        Returns
        -------
        bool
        """
        return isinstance(self, StringKey)


@total_ordering
class NewRow(KeyValue):
    """Key for new document rows in documents that are keyed by the row index.
    Having a dedicated object for these type of keys allows to distinguish
    betweeen primary key values that are None and the new rows.
    """
    def __init__(self, identifier=None):
        """Initialize the optional identifier. This is primarily used for test
        purposes.

        Parameters
        ----------
        identifier: scalar
            Key identifier used for test purposes.
        """
        self.identifier = identifier

    def __eq__(self, other):
        """Objects of this class are equal to any other object of the same
        class.

        Parameters
        ----------
        other: histore.key.base.KeyValue
            Object that this key is compared against.

        Returns
        -------
        bool
        """
        return other.is_new()

    def __lt__(self, other):
        """Objects of this class will always be greater than any other key
        value.

        Parameters
        ----------
        other: histore.key.base.KeyValue
            Object that this key is compared against.

        Returns
        -------
        int
        """
        return False

    def __repr__(self):
        """String representation for the key value.

        Returns
        -------
        string
        """
        return '<NewRow ({})>'.format(self.identifier)


@total_ordering
class NullKey(KeyValue):
    """Object for key values that are None."""
    def __init__(self, identifier=None):
        """Initialize the optional identifier. This is primarily used for test
        purposes.

        Parameters
        ----------
        identifier: scalar
            Key identifier used for test purposes.
        """
        self.identifier = identifier

    def __eq__(self, other):
        """Objects of this class are equal to any other object of the same
        class.

        Parameters
        ----------
        other: histore.key.base.KeyValue
            Object that this key is compared against.

        Returns
        -------
        bool
        """
        return other.is_null()

    def __lt__(self, other):
        """Objects of this class will only be lower than new row keys.

        Parameters
        ----------
        other: histore.key.base.KeyValue
            Object that this key is compared against.

        Returns
        -------
        int
        """
        return True if other.is_new() else False

    def __repr__(self):
        """String representation for the key value.

        Returns
        -------
        string
        """
        return '<Null ({})>'.format(self.identifier)


@total_ordering
class NumberKey(KeyValue):
    """Object for numeric key values."""
    def __init__(self, value):
        """Initialize the actual key value.

        Parameters
        ----------
        value: number
            Numeric key value.
        """
        super(NumberKey, self).__init__(value=value)

    def __eq__(self, other):
        """Use the build-in equality comparison operator when comparing to
        other numeric key values.

        Parameters
        ----------
        other: histore.key.base.KeyValue
            Object that this key is compared against.

        Returns
        -------
        bool
        """
        return self.value == other.value if other.is_number() else False

    def __lt__(self, other):
        """Objects of this class will be lower than any other key type. When
        comparing to a numeric key, use the build in operator to compare the
        key values.

        Parameters
        ----------
        other: histore.key.base.KeyValue
            Object that this key is compared against.

        Returns
        -------
        bool
        """
        return self.value < other.value if other.is_number() else True

    def __repr__(self):
        """String representation for the key value.

        Returns
        -------
        string
        """
        return str(self.value)


@total_ordering
class StringKey(KeyValue):
    """Object for string key values."""
    def __init__(self, value):
        """Initialize the actual key value.

        Parameters
        ----------
        value: string
            String key value.
        """
        super(StringKey, self).__init__(value=value)

    def __eq__(self, other):
        """Use the build-in equality comparison operator when comparing to
        other string key values.

        Parameters
        ----------
        other: histore.key.base.KeyValue
            Object that this key is compared against.

        Returns
        -------
        bool
        """
        return self.value == other.value if other.is_string() else False

    def __lt__(self, other):
        """Objects of this class will be lower than null keys and new row keys.
        They are greater than numeric keys. When comparing to a string key, use
        the build in operator to compare the key values.

        Parameters
        ----------
        other: histore.key.base.KeyValue
            Object that this key is compared against.

        Returns
        -------
        bool
        """
        if other.is_number():
            return False
        elif other.is_string():
            return self.value < other.value
        else:
            return True

    def __repr__(self):
        """String representation for the key value.

        Returns
        -------
        string
        """
        return self.value


# -- Helper Functions ---------------------------------------------------------

def to_key(value):
    """Create an instance of the key value classes for a given scalar value.

    Parameters
    ----------
    value: scalar
        Scalar key value.

    Returns
    -------
    histore.key.base.KeyValue
    """
    if value is None:
        return NullKey()
    elif isinstance(value, KeyValue):
        return value
    elif isinstance(value, int) or isinstance(value, np.integer):
        return NumberKey(value=value)
    elif isinstance(value, float) or isinstance(value, np.floating):
        return NumberKey(value=value)
    elif isinstance(value, str):
        return StringKey(value=value)
    else:
        # By default, use a string representation of the value.
        return StringKey(value=str(value))



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Abstract class for managers that maintain a set of archives."""

from abc import ABCMeta, abstractmethod


class ArchiveManager(metaclass=ABCMeta):  # pragma: no cover
    """The manager for archives implements a factory pattern for creating and
    accessing different types of archives. The type of archive (e.g. persistent
    or volatile is implementaiton dependent. Matdata about each archive is
    maintained in an archive descriptor.
    """
    @abstractmethod
    def archives(self):
        """Get dictionary of archive descriptors. The returned dictionary maps
        archive identifier to their descriptor.

        Returns
        -------
        dict
        """
        raise NotImplementedError()

    def contains(self, identifier):
        """Returns True if an archive with the given identifier exists.

        Parameters
        ----------
        identifier: string
            Unique archive identifier

        Returns
        -------
        bool
        """
        return identifier in self.archives()

    @abstractmethod
    def create(self, name=None, description=None, primary_key=None):
        """Create a new archive object.

        Parameters
        ----------
        name: string, default=None
            Descriptive name that is associated with the archive.
        description: string, default=None
            Optional long description that is associated with the archive.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for rows in the
            archive.

        Returns
        -------
        histore.archive.manager.descriptor.ArchiveDescriptor
        """
        raise NotImplementedError()

    @abstractmethod
    def delete(self, identifier):
        """Delete the archive with the given identifier.

        Parameters
        ----------
        identifier: string
            Unique archive identifier
        """
        raise NotImplementedError()

    @abstractmethod
    def get(self, identifier):
        """Get the archive that is associated with the given identifier. Raises
        a ValueError if the identifier is unknown.

        Parameters
        ----------
        identifier: string
            Unique archive identifier

        Returns
        -------
        histore.archive.vase.Archive

        Raises
        ------
        ValueError
        """
        raise NotImplementedError()

    def list(self):
        """Get the list of descriptors for the maintained archives.

        Returns
        -------
        list(histore.archive.manager.descriptor.ArchiveDescriptor)
        """
        return list(self.archives().values())



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Document wrapper around data frames."""

from histore.document.mem.base import InMemoryDocument

import histore.document.schema as schema
import histore.key.annotate as anno


class DataFrameDocument(InMemoryDocument):
    """Create an in-memory document for a pandas data frame."""
    def __init__(self, df, primary_key=None):
        """Initialize the document columns, rows, and read order. Depending on
        whether a primary key is given or not the document will be keyed by the
        key values or the data frame row index.

        Parameters
        ----------
        df: pandas.DataFrame
            Pandas data frame that is being wrapped by the document class.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for snapshot rows.
        """
        columns = list(df.columns)
        rows = Rows(df=df)
        if primary_key is not None:
            readorder = anno.pk_readorder(
                rows=rows,
                primary_key=schema.column_index(
                    schema=columns,
                    columns=primary_key
                )
            )
        else:
            readorder = anno.rowindex_readorder(index=df.index)
        super(DataFrameDocument, self).__init__(
            columns=columns,
            rows=rows,
            readorder=readorder
        )


class Rows(object):
    """Iterable list for rows in a data frame."""
    def __init__(self, df):
        """Initialize the wrapped data frame.

        Parameters
        ----------
        df: pandas.DataFrame
            Data frame that is wrapped by this class.
        """
        self.df = df

    def __getitem__(self, key):
        """Get data frame row at the given position in the list.

        Parameters
        ----------
        key: int
            List index position for a data frame row.

        Returns
        -------
        pd.Series
        """
        return self.df.iloc[key]

    def __iter__(self):
        """Iterator for rows in the data frame.

        Returns
        -------
        iterator
        """
        return row_stream(self.df)


# -- Helper functions ---------------------------------------------------------

def row_stream(df):
    """Iterator over rows in a data frame.

    Parameters
    ----------
    df: pandas.DataFrame
        Pandas data frame.

    Returns
    -------
    pd.Series
    """
    for _, values in df.iterrows():
        yield values



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Base function to read CSV files as HISTORE documents."""

import csv
import gzip

from histore.document.csv.simple import SimpleCSVDocument
from histore.document.csv.sort import SortedCSVDocument
from histore.document.mem.base import InMemoryDocument

import histore.document.csv.sort as sort
import histore.document.schema as schema
import histore.key.annotate as anno


def read_csv(
    filename, delimiter=',', quotechar='"', compression=None, has_header=True,
    columns=None, primary_key=None, max_size=None
):
    """Read a document from a csv file. If the primary key attributes are given
    the returned document is sorted by the primary key values. Otherwise, the
    original order of rows is kept and each row is assigned a unique index that
    is equal to the row position. Sorts the document by the primary key values
    (if given).

    Parameters
    ----------
    filename: string
        Path to the CSV file that is being read.
    delimiter: string, default=','
        A one-character string used to separate fields.
    quotechar: string, default='"'
        A one-character string used to quote fields containing special
        characters, such as the delimiter or quotechar, or which contain
        new-line characters.
    compression: string, default=None
        String identifier for the file compression format. Will apply to the
        input file and the output file. Currently only 'gzip' compression is
        supported.
    has_header: bool, default=True
        Flag indicating whether the first row of the given file contains the
        column name header.
    columns: list(string), default=None
        List of column names to use. Overrides the names of columns in the file
        if the file has a header.
    primary_key: string or list, default=None
        Column(s) that are used to generate identifier for snapshot rows. The
        columns may be identified by their name or index position. If a string
        in the primary key list refers to a non-unique column name in the file,
        a ValueError is raised.
    max_size: float, default=None
        Maximum size (in MB) of the main-memory buffer for blocks of the CSV
        file that are sorted in main-memory.

    Returns
    -------
    histore.document.base.Document

    Raises
    ------
    ValueError
    """
    # Open the file. Ensure that a valid compression identifier is given.
    if compression is not None and compression != 'gzip':
        raise ValueError("invalid compression format '{}'".format(compression))
    # Raise an error if has_header is False and no list of column names is
    # given. If names are given they have to be all strings.
    if not has_header and columns is None:
        raise ValueError('no column information')
    elif columns is not None:
        if not isinstance(columns, list):
            columns = [columns]
        for col in columns:
            if not isinstance(col, str):
                raise ValueError("invalid column name '{}'".format(col))
    # Open the input file.
    f = open(filename, 'r') if not compression else gzip.open(filename, 'rt')
    try:
        reader = csv.reader(f, delimiter=delimiter, quotechar=quotechar)
        if has_header:
            # Read the header and use the read names as columns (only if no
            # separate list of columns were given).
            header = next(reader)
            if columns is None:
                # Use the header as the list of columns if the list of columns
                # is currently undefined.
                columns = header
        # Create document instance depending on whether a primary key was given
        # or not.
        if primary_key is not None:
            # If a primary key is given we first need to get the index position
            # for the key attributes in the document schema and then sort the
            # input file.
            pk = schema.column_index(schema=columns, columns=primary_key)
            buffer, filenames = sort.split(
                reader=reader,
                sortkey=pk,
                buffer_size=max_size
            )
            if not filenames:
                # If the file fits into main-memory return a sorted in-memory
                # document.
                return InMemoryDocument(
                    columns=columns,
                    rows=buffer,
                    readorder=anno.pk_readorder(rows=buffer, primary_key=pk)
                )
            else:
                # Merge the CSV file blocks and return a sorted CSV document
                # object that wrapps the sorted CSV file.
                return SortedCSVDocument(
                    filename=sort.mergesort(
                        buffer=buffer,
                        filenames=filenames,
                        sortkey=pk
                    ),
                    columns=columns,
                    primary_key=pk
                )
        else:
            # In this case we do not need to sort the document. The document
            # will always be read in original order. Return a file reader for
            # the csv file.
            return SimpleCSVDocument(
                filename=filename,
                columns=columns,
                has_header=has_header,
                delimiter=delimiter,
                quotechar=quotechar,
                compression=compression
            )
    finally:
        # Ensure that the file is colsed.
        f.close()



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""In-memory document created from a Json serialization of a document. The Json
serialization of a document contains the list of columns ('columns') and an
array or arrays for document rows ('data'). The serialization has the optional
primary key element ('primaryKey') that contains a list of key columns.
"""

import jsonschema

from histore.document.mem.base import InMemoryDocument
from histore.document.schema import Column

import histore.document.schema as schema
import histore.key.annotate as anno


"""Json schema for Json document serializations."""

COLUMN_SCHEMA = {
    'type': 'object',
    'properties': {
        'id': {'type': 'number'},
        'name': {'type': 'string'}
    },
    'required': ['id', 'name']
}

DOCUMENT_SCHEMA = {
    'type': 'object',
    'properties': {
        'columns': {
            'type': 'array',
            'items': {'anyOf': [COLUMN_SCHEMA, {'type': 'string'}]}
        },
        'data': {'type': 'array', 'items': {'type': 'array'}},
        'primaryKey': {'type': 'array', 'items': {'type': 'string'}},
        'rowIndex': {'type': 'array', 'items': {'type': 'number'}}
    },
    'required': ['columns', 'data']
}


class JsonDocument(InMemoryDocument):
    """Create an in-memory document from a Json serialization of a document.
    Expects a Json document in the following format:

    {
        "columns": [],
        "data": [[]],
        "primaryKey": ["string" | "int"] | "rowIndex": ["int"]
    }

    Columns are either represented as strings (the column name) or as
    serializations of identifiable columns {"id": "int": "name": "string"}.

    Data is a list of rows, each a list of values.

    For documents that are keyed by a primary key a list of column names is
    expected. For documents that are keyed by a row index the respective list
    of integers is expected. The key elements are optional.
    """
    def __init__(self, doc, validate=True):
        """Create a in-memory document from the given dictionary.

        Parameters
        ----------
        doc: dict
            Json serialization of a document.
        validate: bool default=True
            Validate the given dictionary if the flag is True. Raises a
            ValidationError if the validation fails.

        Raises
        ------
        jsonschema.ValidationError
        """
        self.doc = doc
        if validate:
            jsonschema.validate(instance=doc, schema=DOCUMENT_SCHEMA)
        # Deserialize columns.
        columns = list()
        for obj in doc['columns']:
            if isinstance(obj, dict):
                columns.append(Column(colid=obj['id'], name=obj['name']))
            else:
                columns.append(obj)
        # Get the document rows.
        rows = doc['data']
        # Create the keys for the document rows.
        if 'primaryKey' in doc:
            readorder = anno.pk_readorder(
                rows=rows,
                primary_key=schema.column_index(
                    schema=columns,
                    columns=doc['primaryKey']
                )
            )
        elif 'rowIndex' in doc:
            readorder = anno.rowindex_readorder(index=doc['rowIndex'])
        else:
            readorder = anno.rowindex_readorder(index=list(range(len(rows))))
        # Create instance of the in-memory document.
        super(JsonDocument, self).__init__(
            columns=columns,
            rows=rows,
            readorder=readorder
        )



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Reader for document rows that are accessible in an array in main memory."""

import math

from histore.document.reader import DocumentReader
from histore.document.row import DocumentRow


class InMemoryDocumentReader(DocumentReader):
    """Reader for rows in an array. Reads rows in order defined by the given
    read order.
    """
    def __init__(self, schema, rows, readorder):
        """Initialize the wrapped data frame, the mapping of data frame columns
        to archive schema columns, and the information about row identifier and
        their position in the data frame.

        Parameters
        ----------
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the document rows (based on list position). The schema
            columns provide the unique column identifier that are required to
            generate the returned document rows.
        rows: list
            List of document rows. Each row in the list is an array of the same
            length as the number of columns in the schema.
        readorder: list
            The read order of a document is a sorted list of 3-tuples with
            (list index, key, position). The list is sorted by the row key.
        """
        self.schema = schema
        self.rows = rows
        self.readorder = readorder
        # Maintain the position of the current row for the reader
        self.readindex = 0

    def has_next(self):
        """Test if the reader has more rows to read. If True the next() method
        will return the next row. Otherwise, the next() method will return
        None.

        Returns
        -------
        bool
        """
        return self.readindex < len(self.readorder)

    def next(self):
        """Read the next row in the data frame. Returns None if the end of the
        data frame rows has been reached.

        Returns
        -------
        histore.document.row.DocumentRow
        """
        # Return None if the reader has no more rows.
        if not self.has_next():
            return None
        # Get identifier and position in the data frame for the next row that
        # is being returned.
        rowidx, key, pos = self.readorder[self.readindex]
        self.readindex += 1
        # Create mapping for values in the data frame row to the columns in the
        # archive schema.
        values = dict()
        row = self.rows[rowidx]
        for i in range(len(self.schema)):
            val = row[i]
            try:
                val = None if math.isnan(val) else val
            except TypeError:
                pass
            values[self.schema[i].colid] = val
        return DocumentRow(key=key, pos=pos, values=values)



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Functions for sorting CSV documents in main-memory or using secondary
storage. The functions in this module are inspired by csvsort
(https://github.com/richardpenman/csvsort).

For HISTORE we made a few modifications that allow us to handle CSV files that
can be sorted in main-memory directly instead of writing them to disk after
sorting.
"""

import csv
import heapq
import os
import sys

from tempfile import NamedTemporaryFile as TempFile

from histore.document.base import Document
from histore.document.mem.base import InMemoryDocument
from histore.document.reader import DocumentReader
from histore.document.row import DocumentRow
from histore.key.base import to_key

import histore.config as config


# -- Document class for sorted CSV files --------------------------------------

class SortedCSVDocument(Document):
    """CSV file containing a document that is sorted by one or more of the
    document columns.
    """
    def __init__(self, filename, columns, primary_key):
        """Initialize the object properties.

        Parameters
        ----------
        filename: string
            Path to the CSV file that contains the document.
        primary_key: list
            List of index positions for sort columns.
        """
        self.filename = filename
        self.columns = columns
        self.primary_key = primary_key

    def close(self):
        """Remove the sorted (temporary) file when merging is done."""
        os.remove(self.filename)

    def partial(self, reader):
        """If the document in the CSV file is a partial document we read it
        into a data frame and use the respective data frame document method.
        This requires that partial documents in CSV files fit into main memory.

        Parameters
        ----------
        reader: histore.archive.reader.RowPositionReader
            Reader for row (key, position) tuples from the original
            snapshot version.

        Returns
        -------
        histore.document.base.Document
        """
        rows = list()
        readorder = list()
        with open(self.filename, 'r', newline='') as f:
            csvreader = csv.reader(f)
            for row in csvreader:
                rows.append(row)
                key = rowkey(row, self.primary_key)
                pos = len(readorder)
                readorder.append((pos, key, pos))
        return InMemoryDocument(
            columns=self.columns,
            rows=rows,
            readorder=readorder
        ).partial(reader)

    def reader(self, schema):
        """Get reader for the CSV document.

        Parameters
        ----------
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the column list of this document (corresponding to
            their position in the list). The schema columns provide the unique
            column identifier that are required by the document reader to
            generate document rows.

        Returns
        -------
        histore.document.reader.DocumentReader
        """
        return SortedCSVDocumentReader(
            filename=self.filename,
            schema=schema,
            primary_key=self.primary_key
        )


class SortedCSVDocumentReader(DocumentReader):
    """Document reader for a sorted CSV document."""
    def __init__(self, filename, schema, primary_key):
        """Initialize the CSV reader and the list of column names.

        Parameters
        ----------
        reader: csv.reader
            Reader for the document file. If the file has a header it is
            assumed that the corresponding row has already been read.
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the document rows (based on list position). The schema
            columns provide the unique column identifier that are required to
            generate the returned document rows.
        primary_key: list
            List of index positions for sort columns.
        """
        self.fh = open(filename, 'r', newline='')
        self.reader = csv.reader(self.fh)
        self.schema = schema
        self.primary_key = primary_key
        self._next_row = None
        self._read_index = 0
        self.next()

    def has_next(self):  # pragma: no cover
        """Test if the reader has more rows to read. If this reader has more
        rows the internal row buffer is not empty.

        Returns
        -------
        bool
        """
        return self._next_row is not None

    def next(self):
        """Read the next row in the document. Returns None if the end of the
        document has been reached.

        Returns
        -------
        histore.document.row.DocumentRow
        """
        result = self._next_row
        try:
            row = next(self.reader)
            # If a next row was read ensure that the number of values
            # is the same as the number of columns in the schema.
            if len(row) != len(self.schema):
                self.fh.close()
                raise ValueError('invalid row %d' % (self.reader.line_num))
            # Create document row object for the read values.
            rowpos = self._read_index
            values = dict()
            for i, col in enumerate(self.schema):
                values[col.colid] = row[i]
            key = rowkey(row, self.primary_key)
            self._next_row = DocumentRow(
                key=key,
                pos=rowpos,
                values=values
            )
            self._read_index += 1
        except StopIteration:
            # Close the file if the file is reached.
            self._next_row = None
            self.fh.close()
        # Return the buffered result
        return result


# -- Helper methods for document classes --------------------------------------

def rowkey(row, primary_key):
    """Get the key value for a row.

    Parameters
    ----------
    row: list
        List of cells in a row.
    primary_key: list
        List of index positions for primary key columns.

    Returns
    -------
    scalar or tuple
    """
    if len(primary_key) == 1:
        return to_key(row[primary_key[0]])
    else:
        return tuple([to_key(row[c]) for c in primary_key])


# -- Helper methods for external merge sort -----------------------------------

def decorated_buffer(buffer, sortkey):
    """Iterator for in-memory buffer of data rows.

    Parameters
    ----------
    buffer: list
        Buffer of rows from the input file that were not written to file.
    sortkey: list
        List of index positions for sort columns.
    """
    buffer.sort(key=lambda row: keyvalue(row, sortkey))
    for row in buffer:
        yield keyvalue(row, sortkey), row


def decorated_csv(filename, columns):
    """Iterator for sorted CSV file block. From
    https://github.com/richardpenman/csvsort/blob/master/__init__.py

    Parameters
    ----------
    filename: string
        Temporary CSV file (block) on disk.
    columns: list
        List of index positions for sort columns.
    """
    with open(filename, 'r', newline='') as f:
        for row in csv.reader(f):
            yield keyvalue(row, columns), row


def keyvalue(row, columns):
    """Get the sort key for a given row. From
    https://github.com/richardpenman/csvsort/blob/master/__init__.py

    Parameters
    ----------
    row: list
        List of cell values in a CSV file row.
    columns: list
        List of index positions for sort columns.

    Returns
    -------
    list
    """
    return [row[column] for column in columns]


def mergesort(buffer, filenames, sortkey):
    """2-way merge sort for blocks of a CSV file. Adopted from
    https://github.com/richardpenman/csvsort/blob/master/__init__.py

    Returns name of file contaiing the sorted output.

    Parameters
    ----------
    buffer: list
        Buffer of rows from the input file that were not written to file.
    filename: list
        List of names for sorted temporary files that were generated by the
        CSV split method.
    sortkey: list
        List of index positions for sort columns.

    Returns
    -------
    string
    """
    # Merge the buffer if it contains rows.
    if buffer:
        mergefile = filenames[0]
        with TempFile(delete=False, mode='w', newline='') as f_out:
            writer = csv.writer(f_out)
            files = [
                decorated_buffer(buffer, sortkey),
                decorated_csv(mergefile, sortkey)
            ]
            for _, row in heapq.merge(*files):
                writer.writerow(row)
            filenames = filenames[1:] + [f_out.name]
        os.remove(mergefile)
    while len(filenames) > 1:
        mergefiles, filenames = filenames[:2], filenames[2:]
        with TempFile(delete=False, mode='w', newline='') as f_out:
            writer = csv.writer(f_out)
            files = [decorated_csv(f, sortkey) for f in mergefiles]
            for _, row in heapq.merge(*files):
                writer.writerow(row)
            filenames.append(f_out.name)
        for filename in mergefiles:
            os.remove(filename)
    return filenames[0]


def split(reader, sortkey, buffer_size=None):
    """Split a CSV file into blocks of maximum size. Individual blocks are
    written to temporary files on disk. Only the final buffer is maintained in
    memory. Returns the memory buffer and the list of names for temporary files
    that are created. Adopted from
    https://github.com/richardpenman/csvsort/blob/master/__init__.py

    Returns a tuple of: main-memory buffer for data rows, and list of names for
    temporary files.

    Parameters
    ----------
    reader: csv.reader
        CSV file reader. It is assumed that any header row has been read, i.e,
        the next row in the reader is the first data row of the CSV file.
    sortkey: list
        List of index positions for sort columns.
    buffer_size: float, default=None
        Maximum size of CSV file blocks that are kept in main-memory. If no
        buffer size is given the value from the environment variable
        HISTORE_SORTBUFFER is used.

    Returns
    -------
    list(rows), list(filenames)
    """
    # Ensure that the buffer size is set.
    if buffer_size is None:
        buffer_size = config.SORTBUFFER()
    # Convert buffer size from MB to bytes.
    max_size = buffer_size * 1024 * 1024
    # Split CSV file rows into blocks
    buffer = list()
    current_size = 0
    tmp_filenames = list()
    for row in reader:
        buffer.append(row)
        current_size += sys.getsizeof(row)
        if current_size > max_size:
            # Sort buffer.
            buffer.sort(key=lambda row: keyvalue(row, sortkey))
            # Write buffer to disk.
            with TempFile(delete=False, mode='w', newline='') as f_out:
                writer = csv.writer(f_out)
                for r in buffer:
                    writer.writerow(r)
                # Append file name to result file name list.
                tmp_filenames.append(f_out.name)
            # Clear buffer.
            buffer = list()
            current_size = 0
    return buffer, tmp_filenames



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Helper methods for generating a read order list for documents that are
keyed by a primary key or by list of row index values.
"""

import numpy as np

from histore.key.base import NewRow, NumberKey, to_key


def pk_readorder(rows, primary_key):
    """Create a read order list for a document that derives row identifier from
    cell values using a list of primary key colums.

    Returns a list of 3-tuples with (row position, key, row position). The
    returned list is sorted by the key values.

    Parameters
    ----------
    rows: list
        List of document rows.
    primary_key: list(int)
        List of index positions for primary key columns.

    Returns
    -------
    list

    Raises
    ------
    ValueError
    """
    # For performance reasons we use two different implementations here
    # that distinguish between primary keys with a single column or with
    # multiple columns.
    readorder = list()
    if len(primary_key) == 1:
        keycol = primary_key[0]
        for values in rows:
            key = to_key(values[keycol])
            pos = len(readorder)
            readorder.append((pos, key, pos))
    else:
        for values in rows:
            key = tuple([to_key(values[c]) for c in primary_key])
            pos = len(readorder)
            readorder.append((pos, key, pos))
    return sorted(readorder, key=lambda r: r[1])


def rowindex_readorder(index):
    """Create a read order list from a given list of row index values. The
    index list, for example, corresponds to the row index of a pandas data
    frame.

    Index values are considered as primary key values for document rows. Index
    values that are not positive integers are treated as new rows.

    Returns a list of 3-tuples with (row position, key, row position). The row
    position represents the index position of a row in the given list. The key
    is the value in the index at the poistion (or a new row identifier if the
    original value was negative integer or not of type integer). The returned
    list is sorted by the key values.

    Parameters
    ----------
    index: list
        List of row index values.

    Returns
    -------
    list
    """
    # Ensure that all row identifier are integers. Generate identifier for
    # those rows that have a non-integer index, None, or a negative value
    # as their index. These rows are considered new rows.
    readorder = list()
    for ridx in index:
        pos = len(readorder)
        is_int = isinstance(ridx, int) or isinstance(ridx, np.integer)
        if not is_int or ridx < 0:
            key = NewRow(identifier=pos)
        else:
            key = NumberKey(value=ridx)
        readorder.append((pos, key, pos))
    return sorted(readorder, key=lambda r: r[1])



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Implementation of the document interface for documents that have been read
completely into main memory.
"""

from histore.document.base import Document
from histore.document.mem.reader import InMemoryDocumentReader


class InMemoryDocument(Document):
    """The in-memory document maintains a array of document rows (each as a
    list of cell values) and a sorted row read order as a list of pairs of row
    key and original row position in the document.  These row keys are either
    derived using a primary key (i.e., from the cell values in each row) or
    from the row index of a data frame.
    """
    def __init__(self, columns, rows, readorder):
        """Initialize the object properties. The abstract document class
        maintains a list of row (identifier, position) tuples in the order
        in which rows are being returned by the document reader (i.e., sorted
        by their row key).

        Parameters
        ----------
        columns: list
            List of column names. The number of values in each row is expected
            to be the same as the number of columns and the order of values in
            each row is expected to correspond to their respective column in
            this list.
        rows: list
            List of document rows. Each row in the list is an array of the same
            length as the number of columns in the document.
        readorder: list
            The read order of a document is a sorted list of 3-tuples with
            (list index, key, position). The list is sorted by the row key. The
            list index is the index of a row in the given row list. The key is
            the unique key value for the associated row. The position is the
            position of the row in the document. The list index and the
            position may differ for a row, especially for partial documents.
        """
        super(InMemoryDocument, self).__init__(columns=columns)
        self.rows = rows
        self.readorder = readorder

    def close(self):
        """There are no resources that need to be released by the in-memory
        document.
        """
        pass

    def partial(self, reader):
        """Return a copy of the document where the rows are aligned with the
        position of the corresponding row in the document origin. The original
        row order is accessible via the given row position reader.

        Parameters
        ----------
        reader: histore.archive.reader.RowPositionReader
            Reader for row (key, position) tuples from the original
            snapshot version.

        Returns
        -------
        histore.document.base.Document
        """
        matched_rows, unmatched_rows = list(), list()
        readidx = 0
        orig_row = reader.next()
        max_pos = -1
        while orig_row and readidx < len(self.readorder):
            rowidx, key, pos = self.readorder[readidx]
            if orig_row[0] < key:
                # The original row is not included in the partial document.
                # Get information for next row in the original snapshot.
                orig_row = reader.next()
            elif orig_row[0] > key:
                # The row in the partial document is a new row that was not
                # present in the original document. Add the row identifier to
                # the list of unmatched rows.
                unmatched_rows.append((rowidx, key))
                readidx += 1
            else:
                # Matched row. Adjust position of the row to the position of
                # the row in the original document.
                matched_rows.append((rowidx, key, orig_row[1]))
                orig_row = reader.next()
                readidx += 1
            # Keep track of the maximum position in the original document.
            if orig_row and orig_row[1] > max_pos:
                max_pos = orig_row[1]
        # Finish reading the original document to get the maximum position
        # value.
        while orig_row:
            if orig_row[1] > max_pos:
                max_pos = orig_row[1]
            orig_row = reader.next()
        # Append unmatched rows to the matched rows with adjusted row position.
        for rowidx, key in unmatched_rows:
            max_pos += 1
            matched_rows.append((rowidx, key, max_pos))
        # Append remaining rows in the partial document to the matched rows
        # list with adjusted position.
        while readidx < len(self.readorder):
            max_pos += 1
            rowidx, key, pos = self.readorder[readidx]
            matched_rows.append((rowidx, key, max_pos))
            readidx += 1
        # Ensure that all document rows have been matched (currently only done
        # by comparing the length of two lists). This is a sanity check in case
        # of a bug in the above code.
        if len(self.readorder) != len(matched_rows):  # pragma: no cover
            raise RuntimeError('did not match all rows')
        # Initialize the superclass with adjusted row positions.
        matched_rows.sort(key=lambda r: r[1])
        return InMemoryDocument(
            columns=self.columns,
            rows=self.rows,
            readorder=matched_rows
        )

    def reader(self, schema):
        """Get reader for document rows ordered by their row key.

        Parameters
        ----------
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the column list of this document (corresponding to
            their position in the list). The schema columns provide the unique
            column identifier that are required by the document reader to
            generate document rows. An error is raised if the number of
            elements in the schema does not match the number of columns in the
            data frame.

        Returns
        -------
        histore.document.mem.reader.InMemoryDocumentReader

        Raises
        ------
        ValueError
        """
        # Raise an error if the number of elements in the schema does not match
        # the number of columns in the document.
        if len(self.columns) != len(schema):
            raise ValueError('invalid schema for data frame')
        return InMemoryDocumentReader(
            schema=schema,
            rows=self.rows,
            readorder=self.readorder
        )



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Base classes for maintaining metadata about the archives that are controlled
by an archive manager.
"""

import jsonschema

import histore.util as util


"""Json schema for archive descriptors."""

DESCRIPTOR_SCHEMA = {
    'type': 'object',
    'properties': {
        'id': {'type': 'string'},
        'name': {'type': 'string'},
        'description': {'type': 'string'},
        'primaryKey': {
            'type': 'array',
            'items': {'type': 'string'}
        }
    },
    'required': ['id']
}


class ArchiveDescriptor(object):
    """Wrapper around an archive descriptor dictionary object. This class
    provides access to descriptor property values and their defaults.
    """
    def __init__(self, doc, validate=True):
        """Initialize the dictionary containing the archive descriptor.
        Validates the document against the descriptor schema if the validate
        flag is True. Raises a ValidationError if validation fails.

        Parameters
        ----------
        doc: dict
            Dictionary containing an archive descriptor. The dictionary should
            adhere to the defined descriptor schema.
        validate: bool default=True
            Validate the given dictionary against the descriptor schema if this
            flag is True. Raises a ValidationError if the validation fails.

        Raises
        ------
        jsonschema.ValidationError
        """
        self.doc = doc
        if validate:
            jsonschema.validate(instance=doc, schema=DESCRIPTOR_SCHEMA)

    @staticmethod
    def create(name=None, description=None, primary_key=None):
        """Create a new archive descriptor object.

        Parameters
        ----------
        name: string, default=None
            Descriptive name that is associated with the archive.
        description: string, default=None
            Optional long description that is associated with the archive.
        primary_key: string or list, default=None
            Column(s) that are used to generate identifier for rows in the
            archive.

        Returns
        -------
        histore.archive.manager.base.ArchiveDescriptor
        """
        # Ensure that the primary key is a list
        if primary_key is not None and not isinstance(primary_key, list):
            primary_key = [primary_key]
        # Create a unique identifier for the new archive.
        identifier = util.get_unique_identifier()
        # Create the archive descriptor.
        doc = {'id': identifier}
        if name is not None:
            doc['name'] = name
        if description is not None:
            doc['description'] = description
        if primary_key is not None:
            doc['primaryKey'] = primary_key
        return ArchiveDescriptor(doc)

    def description(self):
        """Get archive description. If the value is not set in the descriptor
        an empty string is returned as default.

        Returns
        -------
        string
        """
        return self.doc.get('description', '')

    def identifier(self):
        """Get the unique archive identifier value.

        Returns
        -------
        string
        """
        return self.doc['id']

    def name(self):
        """Get the archive name. If the value is not set in the descriptor the
        identifier is returned as default.

        Returns
        -------
        string
        """
        return self.doc.get('name', self.identifier())

    def primary_key(self):
        """Get list of primary key attributes.

        Returns
        -------
        list(string)
        """
        return self.doc.get('primaryKey')



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Functions and classes to read CSV files as HISTORE documents."""

import csv
import gzip
import pandas as pd

from histore.document.base import Document
from histore.document.mem.dataframe import DataFrameDocument
from histore.document.reader import DocumentReader
from histore.document.row import DocumentRow
from histore.key.base import NumberKey


class SimpleCSVDocument(Document):
    """CSV file containing a document that is keyed by the row position and
    not a primary key (refered to as an un-keyed document). Returns rows in
    document order with the row identifier being the line number (starting at
    zero and ignoring an optional column header).
    """
    def __init__(
        self, filename, columns, delimiter=',', quotechar='"',
        compression=None, has_header=True
    ):
        """Initialize the object properties. The abstract document class
        maintains a list of column names in the document schema. Columns may
        either be represented by strings or by instances of the Column class.

        Parameters
        ----------
        filename: string
            Path to the CSV file that contains the document.
        columns: list
            List of column names. The number of values in each document row is
            expected to be the same as the number of columns and the order of
            values in each row is expected to correspond to their respective
            column in this list.
        delimiter: string, default=','
            A one-character string used to separate fields.
        quotechar: string, default='"'
            A one-character string used to quote fields containing special
            characters, such as the delimiter or quotechar, or which contain
            new-line characters.
        compression: string, default=None
            String identifier for the file compression format. Currently only
            'gzip' compression is supported.
        has_header: bool, default=True
            Flag indicating whether the first row of the given file contains
            column names. The header row will be ignored by the reader in favor
            of the given column names.
        """
        self.filename = filename
        self.columns = columns
        self.delimiter = delimiter
        self.quotechar = quotechar
        self.compression = compression
        self.has_header = has_header

    def close(self):
        """The un-keyed CSV document reader has no additional resources that
        need to be released.
        """
        pass  # pragma: no cover

    def open(self):
        """Open the associated file for reading.

        Returns
        -------
        FileObject
        """
        if self.compression == 'gzip':
            return gzip.open(self.filename, 'rt')
        else:
            return open(self.filename, 'r')

    def partial(self, reader):
        """If the document in the CSV file is a partial document we read it
        into a data frame and use the respective data frame document method.
        This requires that partial documents in CSV files fit into main memory.

        Parameters
        ----------
        reader: histore.archive.reader.RowPositionReader
            Reader for row (key, position) tuples from the original
            snapshot version.

        Returns
        -------
        histore.document.base.Document
        """
        df = pd.read_csv(
            self.filename,
            delimiter=self.delimiter,
            quotechar=self.quotechar,
            header=0 if self.has_header else None,
            names=self.columns,
            compression=self.compression
        )
        return DataFrameDocument(df).partial(reader)

    def reader(self, schema):
        """Get reader for data frame rows ordered by their row identifier. In
        a partial document the row positions that are returned by the reader
        are aligned with the positions of the corresponding rows in the
        document of origin.

        Parameters
        ----------
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the column list of this document (corresponding to
            their position in the list). The schema columns provide the unique
            column identifier that are required by the document reader to
            generate document rows. An error is raised if the number of
            elements in the schema does not match the number of columns in the
            data frame.

        Returns
        -------
        histore.document.reader.DocumentReader

        Raises
        ------
        ValueError
        """
        # Open the input file and skip the header.
        f = self.open()
        try:
            reader = csv.reader(
                f,
                delimiter=self.delimiter,
                quotechar=self.quotechar
            )
            if self.has_header:
                # Skip the column header (if present).
                next(reader)
        except IOError as ex:  # pragma: no cover
            # Ensure that the file is closed in case of an error.
            f.close()
            raise ex
        return SimpleCSVDocumentReader(reader=reader, schema=schema, fh=f)


class SimpleCSVDocumentReader(DocumentReader):
    """Document reader for un-keyed CSV documents."""
    def __init__(self, reader, schema, fh):
        """Initialize the CSV reader and the list of column names.

        Parameters
        ----------
        reader: csv.reader
            Reader for the document file. If the file has a header it is
            assumed that the corresponding row has already been read.
        schema: list(histore.document.schema.Column)
            List of columns in the document schema. Each column corresponds to
            a column in the document rows (based on list position). The schema
            columns provide the unique column identifier that are required to
            generate the returned document rows.
        fh: FileObject
            Handle for the file that is being read. Required to be able to
            properly close the file when done with reading.
        """
        self.reader = reader
        self.schema = schema
        self.fh = fh
        self._next_row = None
        self._row_id = 0
        self.next()

    def has_next(self):  # pragma: no cover
        """Test if the reader has more rows to read. If this reader has more
        rows the internal row buffer is not empty.

        Returns
        -------
        bool
        """
        return self._next_row is not None

    def next(self):
        """Read the next row in the document. Returns None if the end of the
        document has been reached.

        Returns
        -------
        histore.document.row.DocumentRow
        """
        result = self._next_row
        try:
            row = next(self.reader)
            # If a next row was read ensure that the number of values
            # is the same as the number of columns in the schema.
            if len(row) != len(self.schema):
                self.fh.close()
                raise ValueError('invalid row %d' % (self.reader.line_num))
            # Create document row object for the read values.
            rowid = self._row_id
            values = dict()
            for i, col in enumerate(self.schema):
                values[col.colid] = row[i]
            self._next_row = DocumentRow(
                key=NumberKey(rowid),
                pos=rowid,
                values=values
            )
            self._row_id += 1
        except StopIteration:
            # Close the file if the file is reached.
            self._next_row = None
            self.fh.close()
        # Return the buffered result
        return result



# This file is part of the History Store (histore).
#
# Copyright (C) 2018-2020 New York University.
#
# The History Store (histore) is released under the Revised BSD License. See
# file LICENSE for full license details.

"""Helper classes for unit tests."""


class ListReader(object):
    """Implementation of the reader interface. Returns values from a given list
    until the list is empty.
    """
    def __init__(self, values):
        """Initialize the list of returned values.

        Parameters
        ----------
        values: list
            List of returned values.
        """
        # Reverse the list of values to be able to remove values in original
        # order using pop()
        self.values = values[::-1]

    def has_next(self):
        """Returns True if the internal list is not empty (i.e., there are
        values left to be returned by the next() method).

        Returns
        -------
        bool
        """
        return self.values

    def next(self):
        """Return the next value in the list.

        Returns
        -------
        object
        """
        if self.has_next():
            return self.values.pop()
